# -*- coding: utf-8 -*-
"""
helpers.text_utils

Utility text and string processing functions

* created: 2013-10-27 kevin chan <kefin@makedostudio.com>
* updated: 2014-11-21 kchan
"""

from __future__ import (absolute_import, unicode_literals, print_function)

import six
import re
import string


default_encoding = "utf-8"

def uprint(data, encoding=default_encoding):
    """
    Print unicode output to stdout.

    :param data: data to print.
    :param encoding: data encoding (default is `utf-8`).
    """
    try:
        print(data.encode(encoding))
    except AttributeError:
        print(data)


# utility string functions

def trim(s):
    """Trim white space from beginning and end of string."""
    return s.lstrip().rstrip()


CR = '\n'

def check_eos(s, cr=CR):
    """
    Check end of string and make sure there's a return.

    :param s: text
    :param cr: EOL character or string
    :returns: text wth EOL appended
    """
    try:
        if not s.endswith(cr):
            s += cr
    except (TypeError, AttributeError):
        pass
    return s


# string test functions

def has_digits(s):
    """
    Test if string has digits.

    :param s: string
    :returns: number of digits in string
    """
    return len(set(s) & set(string.digits))

def has_alpha(s):
    """
    Test if string has alphabets.

    :param s: string
    :returns: number of letters in string
    """
    return len(set(s) & set(string.letters))

def has_alphanum(s):
    """
    Test if string has alphabets and digits.

    :param s: string
    :returns: number of letters and digits in string
    """
    alphanum = set(string.letters + string.digits)
    return len(set(s) & alphanum)


def tidy_txt(txt):
    """
    Utility function to tidy up text block by compressing multiple
    blank lines.
    * multiple (>2) blank lines are reduced to 2.

    :param txt: text block
    :returns: tidied up text block with multiple blank lines removed.
    """
    if isinstance(txt, six.text_type):
        lines = unicode(txt).split('\n')
        s = []
        run = 0
        for line in lines:
            txt = line.strip()
            if len(txt) == 0:
                if run < 1:
                    s.append('')
                run += 1
            else:
                run = 0
                s.append(txt)
        txt = '\n'.join(s)
    return txt


# utility functions to convert names to/from CamelCase

def to_camel_case(name):
    """
    Convert name to CamelCase.
    * does not do any sanity checking (assumes name
      is somewhat already alphanumeric).
    """
    delete_chars = """'":;,~!@#$%^&*()_+-`=<>./?\\|"""
    result = []
    prev = ' '
    for c in name:
        if not prev.isalnum():
            c = c.upper()
        else:
            c = c.lower()
        result.append(c)
        prev = c
    return "".join(result).translate(string.maketrans(' -','__'), delete_chars)


# perform substitution on a chunk of text

# default id pattern: ${VARIABLE}
IdPattern = r'\$\{([a-z_][a-z0-9_]*)\}'
IdRegexp = re.compile(IdPattern, re.I)

def substitute(txt, context, pattern=None):
    """
    Perform variable substitution on a chunk of text.
    * returns None if input text is None.
    * default var pattern is ${var}

    Parameters:
    * txt is text or template to perform substitution on
    * context is dict of key/value pairs or callable to retrieve values
    * pattern is regexp pattern or compiled regexp object to perform match
    """
    if txt is None:
        return None
    if pattern is None:
        regexp = IdRegexp
    elif isinstance(pattern, basestring):
        regexp = re.compile(pattern, re.I)
    else:
        regexp = pattern
    if callable(context):
        getval = context
    else:
        if context is None:
            context = {}
        getval = lambda kw: context.get(kw, '')
    done = False
    while not done:
        matches = regexp.findall(txt)
        if len(matches) > 0:
            txt = regexp.sub(lambda m: getval(m.group(1)), txt)
        else:
            done = True
    return txt


# simple string substitution function

def subs(template, context):
    """
    Perform simple string substitutions using string template

    Example:

    caption_template = '<div class="%(caption_css_class)s">%(caption)s</div>'
    caption = subs(caption_template,
                  {
                      'caption_css_class': 'image-caption',
                      'caption': 'test'
                  })

    """
    result = ''
    try:
        assert template is not None
        result = template % context
    except (AssertionError, TypeError):
        pass
    return result


# from: http://code.activestate.com/recipes/466341-guaranteed-conversion-to-unicode-or-byte-string/
# Recipe 466341 (r1): Guaranteed conversion to unicode or byte string

def safe_unicode(obj, *args):
    """ return the unicode representation of obj """
    try:
        return unicode(obj, *args)
    except UnicodeDecodeError:
        # obj is byte string
        ascii_text = str(obj).encode('string_escape')
        return unicode(ascii_text)

def safe_str(obj):
    """ return the byte string representation of obj """
    try:
        return str(obj)
    except UnicodeEncodeError:
        # obj is unicode
        return unicode(obj).encode('unicode_escape')


# utility function to truncate text

def truncate_chars(data, maxlen):
    """
    Truncate string to at most ``maxlen``, including elipsis.
    * uses django.utils.text Truncator class.

    :param data: string to truncate
    :param maxlen: length to truncate to
    :returns: string (truncated if necessary)
    """
    from django.utils.text import Truncator
    if isinstance(data, six.text_type) and len(data) > maxlen:
        truncator = Truncator(data)
        data = truncator.chars(maxlen)
    return data



# -*- coding: utf-8 -*-
"""
garage.session

Helper functions to set/get session data.

* created: 2011-02-15 Kevin Chan <kefin@makedostudio.com>
* updated: 2014-11-21 kchan
"""

from __future__ import (absolute_import, unicode_literals)


# get/set session vars

def set_session_var(request, skey, sval):
    """Set key-value in session cookie."""
    try:
        request.session[skey] = sval
    except (TypeError, AttributeError):
        pass


def get_session_var(request, skey, default=None):
    """Get value from session cookie."""
    try:
        return request.session.get(skey, default)
    except (TypeError, AttributeError):
        return default



# -*- coding: utf-8 -*-
"""
garage.slugify

Functions to create slugs.

* created: 2011-02-15 Kevin Chan <kefin@makedostudio.com>
* updated: 2014-11-21 kchan
"""

from __future__ import (absolute_import, unicode_literals)

import re
import string
from unicodedata import normalize


# general slugify function

SlugDeleteChars = """'"‘’“”:;,~!@#$%^*()_+`=<>./?\\|—–"""
SubstChar = "-"

def strip_accents(s):
    """Strip accents from string and return ascii version."""
    return unicode(normalize('NFKD', unicode(s)).encode('ASCII', 'ignore'))


def slugify(s, delete_chars=SlugDeleteChars, subst_char=SubstChar):
    """
    Convert (unicode) string to slug.
    """
    from garage.text_utils import safe_unicode
    from garage.html_utils import strip_tags, unescape
    def convert_unwanted_chars(txt):
        converted = []
        for ch in txt:
            if ch in delete_chars:
                ch = subst_char
            converted.append(ch)
        return ''.join(converted)

    try:
        s = s.decode("utf-8")
    except UnicodeEncodeError:
        s = safe_unicode(s)
    s = s.strip("\r\n")
    s = s.replace("\n", " ")
    s = strip_accents(s)
    s = strip_tags(unescape(s))
    s = re.sub(r"['’]s", 's', s)
    s = re.sub(r'([0-9\.]+)%', '\\1-percent', s)
    s = s.replace("&amp;", " and ")
    s = s.replace("&", " and ")
    s = s.replace("/", " ")
    s = s.replace(" ", "-")
    s = s.replace("_", "-")
    s = convert_unwanted_chars(s)
    s = re.sub(r'\.\.+', '.', s)
    s = re.sub(r'--+', '-', s)
    s = s.strip('.')
    s = s.strip('-')
    s = s.lower()
    return s


# function to generate unique slugs for django model entries

# default slug spearators
SLUG_SEPARATOR = '-'
SLUG_ITERATION_SEPARATOR = '--'

def get_slug_separator():
    from garage import get_setting
    return get_setting('SLUG_SEPARATOR', SLUG_SEPARATOR)

def get_slug_iteration_separator():
    """
    The slug iteration separator is used to mark entries with the same slug base.

    e.g. article (first entry)
         article--2 (second entry)
    """
    from garage import get_setting
    return get_setting('SLUG_ITERATION_SEPARATOR', SLUG_ITERATION_SEPARATOR)


slug_pat = r'^(.+)%s(\d+)$'
slug_regex = None

def get_slug_base(s, slug_iteration_separator=None):
    """
    Return slug minus the slug_iteration_separator + 'n' sufix.

    * Example: 'article--2' will return 'article' if
    slug_iteration_separator is '--'.
    """
    global slug_regex
    if not slug_iteration_separator:
        slug_iteration_separator = get_slug_iteration_separator()
    if slug_regex is None:
        slug_regex = re.compile(slug_pat % slug_iteration_separator, re.I)
    m = slug_regex.match(s)
    if m:
        return m.group(1)
    return s


def slug_creation_error(msg=None):
    from django.core.exceptions import ValidationError
    if msg is None:
        msg = 'Unable to create slug.'
    raise ValidationError(msg)


def get_unique_slug(instance, slug_field, queryset=None, slug_base=None,
                    prefix=None, suffix=None, slug_separator=None):
    """
    Helper function to generate unique slug for model entries.

    * object instance must already exist.
    * unqiue slug has the format: prefix-slug-n
    * examples:

    if prefix is '20110311-', generated unique slugs for identically
    slugged articles will produce:

    article -> 20110311-article
    article -> 20110211-article--2
    article -> 20110211-article--3
    etc.

    :param instance: object instance
    :param slug_base: slug to use as base for unique slug
    :param slug_field: unique slug field name for queries to test uniqueness
    :param queryset: queryset to use for db access (optional)
    :param prefix: prefix to prepend to unique slug before doing query
    :param suffix: suffix to prepend to unique slug before doing query
    :param slug_separator: string to separate last part of the slug
           from the base (default: SLUG_ITERATION_SEPARATOR)
    :returns: (unque_slug, unique_slug_without_prefix)
    """
    if not prefix:
        prefix = ''
    if not suffix:
        suffix = ''
    if not slug_separator:
        slug_separator = get_slug_iteration_separator()
    try:
        if queryset is None:
            queryset = instance.__class__._default_manager.all()
        if instance.pk:
            queryset = queryset.exclude(pk=instance.pk)
        if slug_base is None:
            try:
                slug_base = getattr(instance, slug_field)
            except AttributeError:
                slug_base = 'entry'
        slug = None
        num = ''
        next = 1
        while 1:
            slug = '%s%s%s' % (slug_base, num, suffix)
            unique_slug = '%s%s' % (prefix, slug)
            if not queryset.filter(**{slug_field: unique_slug}):
                return (unique_slug, slug)
            next += 1
            num = '%s%d' % (slug_separator, next)
    except (AttributeError, TypeError):
        slug_creation_error()


def create_unique_slug(obj, slug_field=None):
    """
    Create simple unique slug for object instance.

    * if slug_field is not supplied, assume it's called 'slug'.

    :param instance: model object instance
    :returns: unique slug
    """
    if not slug_field:
        slug_field = 'slug'
    try:
        sbase = get_slug_base(getattr(obj, slug_field))
    except (AttributeError, TypeError):
        sbase = None
    slugs = get_unique_slug(obj, slug_field=slug_field, slug_base=sbase)
    return slugs[0]



# unique slug function
# from: http://djangosnippets.org/snippets/690/
#
# def unique_slugify(instance, value, slug_field_name='slug', queryset=None,
#                  slug_separator='-'):
#   """
#   Calculates and stores a unique slug of ``value`` for an instance.
#
#   ``slug_field_name`` should be a string matching the name of the field to
#   store the slug in (and the field to check against for uniqueness).
#
#   ``queryset`` usually doesn't need to be explicitly provided - it'll default
#   to using the ``.all()`` queryset from the model's default manager.
#   """
#   slug_field = instance._meta.get_field(slug_field_name)
#
#   slug = getattr(instance, slug_field.attname)
#   slug_len = slug_field.max_length
#
#   # Sort out the initial slug, limiting its length if necessary.
#   slug = slugify(value)
#   if slug_len:
#       slug = slug[:slug_len]
#   slug = _slug_strip(slug, slug_separator)
#   original_slug = slug
#
#   # Create the queryset if one wasn't explicitly provided and exclude the
#   # current instance from the queryset.
#   if queryset is None:
#       queryset = instance.__class__._default_manager.all()
#   if instance.pk:
#       queryset = queryset.exclude(pk=instance.pk)
#
#   # Find a unique slug. If one matches, at '-2' to the end and try again
#   # (then '-3', etc).
#   next = 2
#   while not slug or queryset.filter(**{slug_field_name: slug}):
#       slug = original_slug
#       end = '%s%s' % (slug_separator, next)
#       if slug_len and len(slug) + len(end) > slug_len:
#           slug = slug[:slug_len-len(end)]
#           slug = _slug_strip(slug, slug_separator)
#       slug = '%s%s' % (slug, end)
#       next += 1
#
#   setattr(instance, slug_field.attname, slug)
#
#
# def _slug_strip(value, separator='-'):
#   """
#   Cleans up a slug by removing slug separator characters that occur at the
#   beginning or end of a slug.
#
#   If an alternate separator is used, it will also replace any instances of
#   the default '-' separator with the new separator.
#   """
#   separator = separator or ''
#   if separator == '-' or not separator:
#       re_sep = '-'
#   else:
#       re_sep = '(?:-|%s)' % re.escape(separator)
#   # Remove multiple instances and if an alternate separator is provided,
#   # replace the default '-' separator.
#   if separator != re_sep:
#       value = re.sub('%s+' % re_sep, separator, value)
#   # Remove separator from the beginning and end of the slug.
#   if separator:
#       if separator != '-':
#           re_sep = re.escape(separator)
#       value = re.sub(r'^%s+|%s+$' % (re_sep, re_sep), '', value)
#   return value

#######################################################################



# -*- coding: utf-8 -*-
"""
garage.help_text

Helper function to retrieve help text for backend admin form views.

* created: 2011-03-18 Kevin Chan <kefin@makedostudio.com>
* updated: 2014-11-21 kchan
"""

from __future__ import (absolute_import, unicode_literals)

import warnings

# issue deprecation warning
# * This module is here for compatibility with old imports but will be
#   phased out in the next minor version.
warnings.warn('The help_text module will be deprecated in version 0.2.1. ',
              DeprecationWarning)


# maintain a help text registry for django models

HELP_TEXT_REGISTRY = {}

def register_help_text_dictionary(module, dictionary):
    HELP_TEXT_REGISTRY[module] = dictionary


def unregister_help_text_dictionary(module):
    try:
        d = HELP_TEXT_REGISTRY.get(module)
        del HELP_TEXT_REGISTRY[module]
        return d
    except (AttributeError, KeyError):
        return None


def get_help_text_registry(module=None):
    if module:
        return HELP_TEXT_REGISTRY.get(module)
    return HELP_TEXT_REGISTRY


def get_help_text(module, model, field, default_dict={}):
    """
    Legacy function for compatiblity with old projects using the
    `help_text` module.

    Get help text for model and field in module help registry.

    """
    for d in [get_help_text_registry(module), default_dict]:
        try:
            txt = d[model].get(field)
            if txt:
                return txt
        except (TypeError, KeyError):
            pass
    return ''



# -*- coding: utf-8 -*-
"""
garage.cache

Cache helpers.

* created: 2011-03-14 Kevin Chan <kefin@makedostudio.com>
* updated: 2014-11-21 kchan
"""

from __future__ import (absolute_import, unicode_literals)

import six
import functools
import hashlib

from django.core.cache import cache


# cache helpers
# * uses django caching backend

def s2hex(s):
    """
    Convert any string to a hex-digit hash for use as cache key.
    * uses MD5 hash from hashlib

    :param s: string to hash
    :returns: hash of string as hex digits
    """
    from garage.text_utils import safe_str
    try:
        h = hashlib.md5(s).hexdigest()
    except UnicodeEncodeError:
        h = hashlib.md5(safe_str(s)).hexdigest()
    except:
        h = hashlib.md5(repr(s)).hexdigest()
    return h


CACHE_KEY_SEPARATOR = '/'

def cache_key(name, *args, **kwargs):
    """
    helper function to calculate cache key.
    * accepts the following keyword parameters:
      key_separator -- separator string to use when concatenating args
      prefix -- prefix to prepend to key (after hashing)

    :param name: text to create key hash
    :param args: list of strings to join to "name"
    :param kwargs: keyword arguments (see code below for keywords)
    :returns: hashed key
    """
    key_separator = kwargs.get('key_separator', CACHE_KEY_SEPARATOR)
    prefix = kwargs.get('prefix')
    if not hasattr(name, '__iter__'):
        name = [name]
    if len(args) > 0:
        name.extend(args)
    key = s2hex(key_separator.join(name))
    if prefix is not None:
        key = '%s%s' % (prefix, key)
    return key

def create_cache_key(name, *args, **kwargs):
    """
    Same as cache_key (for compatibility with legacy function).
    """
    return cache_key(name, *args, **kwargs)


DEFAULT_TIMEOUT = 1800

def cache_data(key=None, timeout=DEFAULT_TIMEOUT):
    """
    Decorator to cache objects.
    * see: http://djangosnippets.org/snippets/492/
    """
    def decorator(f):
        @functools.wraps(f)
        def _cache_controller(*args, **kwargs):
            if key is None:
                k = f.__name__
            elif isinstance(key, six.string_types):
                k = key
            elif callable(key):
                k = key(*args, **kwargs)
            result = cache.get(k)
            if result is None:
                result = f(*args, **kwargs)
                cache.set(k, result, timeout)
            return result
        return _cache_controller
    return decorator


def delete_cache(key):
    """
    Delete cached object.

    :param key: key to retrieve data from cache
    :returns: True if cached data is found and deleted else False
    """
    if cache.get(key):
        cache.set(key, None, 0)
        result = True
    else:
        result = False
    return result



# -*- coding: utf-8 -*-
"""
garage.datetime

Datetime and timezone utility functions.

* created: 2013-04-15 Kevin Chan <kefin@makedostudio.com>
* updated: 2014-11-21 kchan
"""

from __future__ import (absolute_import, unicode_literals)

import pytz


def get_local_tz():
    """
    Get local timezone.

    :returns: local timezone object
    """
    from django.utils.timezone import get_current_timezone_name
    from garage import get_setting
    local_timezone = get_setting('TIME_ZONE')
    if not local_timezone:
        local_timezone = get_current_timezone_name()
    local_tz = pytz.timezone(local_timezone)
    return local_tz


def convert_datetime(dt, from_tz, to_tz):
    """
    Convert datetime object from one timezone to another.

    How to use:
    loc_dt = datetime(2013, 4, 15, 22, 30, 30)
    from_tz = pytz.timezone('America/Los_Angeles')
    to_tz = pytz.utc
    utc_dt = convert_datetime(loc_dt, from_tz, to_tz)

    :param dt: datetime object
    :param from_tz: local timezone object
    :param to_tz: timezone object to convert datetime to
    :returns: converted datetime object (in target timezone)
    """
    try:
        t = from_tz.localize(dt)
        src_dt = from_tz.normalize(t)
    except (AttributeError, ValueError):
        src_dt = dt
    try:
        dst_dt = src_dt.astimezone(to_tz)
    except AttributeError:
        dst_dt = dt
    return dst_dt


def local_to_utc(dt, local_tz=None):
    """
    Convert datetime object from local time to UTC.
    * if local_tz is not specified, get timezone from django settings.

    How to use:
    local_dt = datetime(2013, 4, 15, 22, 30, 30)
    utc_dt = local_to_utc(local_dt)

    Reference link:
    * pytz utc conversion
    * http://stackoverflow.com/questions/1357711/pytz-utc-conversion

    :param dt: datetime object
    :param local_tz: use this local timezone object
    :returns: datetime object converted to UTC time
    """
    if local_tz is None:
        local_tz = get_local_tz()
    return convert_datetime(dt, local_tz, pytz.utc)


def utc_to_local(dt, local_tz=None):
    """
    Convert datetime object from UTC to local time.
    * if local_tz is not specified, get timezone from django settings.
    
    How to use:
    dt = datetime(2013, 4, 15, 22, 30, 30)
    utc_dt = dt.replace(tzinfo=pytz.utc)
    loc_dt = utc_to_local(utc_dt)

    :param dt: datetime object
    :param local_tz: use this local timezone object as target
    :returns: datetime object converted to target local timezone
    """
    if local_tz is None:
        local_tz = get_local_tz()
    return convert_datetime(dt, pytz.utc, local_tz)



# -*- coding: utf-8 -*-
"""
garage.utils

Utility functions

* created: 2008-08-11 kevin chan <kefin@makedostudio.com>
* updated: 2014-11-21 kchan
"""

from __future__ import (absolute_import, unicode_literals)

import six
import os
import sys
import re
import hashlib
import base64
import codecs
import pickle
import yaml

try:
    from yaml import (
        CLoader as Loader,
        CDumper as Dumper
    )
except ImportError:
    from yaml import Loader, Dumper


def get_instance(module, class_name, *args, **kwargs):
    """
    Return an instance of the object based on
    module name and class name.

    :param module: module name (e.g. garage.utils)
    :param class_name: name of class defined in module (e.g. DataObject)
    :param args, kwargs: args and kwargs to supply to class when
        creating instance
    :returns: object instance of class
    """
    if not module in sys.modules:
        __import__(module)
    f = getattr(sys.modules[module], class_name)
    obj = f(*args, **kwargs)
    return obj


default_encoding = "utf-8"

def get_file_contents(path, encoding=default_encoding):
    """
    Load text file from file system and return content as string.
    * default encoding is utf-8

    :param path: path to file on local file system
    :param encoding: encoding for file date (default: utf-8)
    :returns: file data or None is file cannot be read
    """
    try:
        assert path is not None and os.path.isfile(path)
        file_obj = codecs.open(path, "r", encoding)
        data = file_obj.read()
        file_obj.close()
    except (TypeError, AssertionError, IOError):
        data = None
    return data


def write_file(path, data, encoding=default_encoding):
    """
    Write text to local file system.

    :param path: path to file to write to on local file system
    :param encoding: encoding for file date (default: utf-8)
    :returns: True if successful else False
    """
    try:
        the_file = open(path, 'wb')
        the_file.write(data.encode(encoding))
        the_file.close()
        return True
    except IOError:
        return False


def make_dir(path):
    """
    Make sure path exists by create directories
    * path should be directory path
      (example: /home/veryloopy/www/app/content/articles/archives/)

    :param path: path to directory
    :returns: True if directory exists else False
    """
    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except OSError:
            pass
    if os.path.exists(path):
        return True
    else:
        return False


# YAML utilities

# yaml usage:
# data = load(stream, Loader=Loader)
# output = dump(data, Dumper=Dumper)

def load_yaml(data):
    """
    Parse yaml data.

    :param data: YAML-formatted data
    :returns: loaded data structure
    """
    return yaml.load(data, Loader=Loader)


def load_yaml_docs(data):
    """
    Parse a series of documents embedded in a YAML file.

    * documents are delimited by '---' in the file

    :param data: YAML-formatted data
    :returns: generator object with loaded data
    """
    return yaml.load_all(data, Loader=Loader)


def dump_yaml(data, **opts):
    """
    Dump data structure in yaml format.

    example usage:
    print(dump_yaml(y, explicit_start=True, default_flow_style=False))

    :param data: data structure
    :param opts: optional parameters for yaml engine
    :returns: YAML-formatted `basestring` for output
    """
    return yaml.dump(data, Dumper=Dumper, **opts)


def sha1hash(s):
    """
    Calculate sha1 hash in hex for string.
    """
    from garage.text_utils import safe_str
    try:
        return hashlib.sha1(s).hexdigest()
    except UnicodeEncodeError:
        return hashlib.sha1(safe_str(s)).hexdigest()
    except:
        return hashlib.sha1(repr(s)).hexdigest()


# encode/decode functions
# * note: encode_sdata and decode_sdata do not perform any sort of
#   encryption

def encode_sdata(data):
    """
    Encode data (dict) using pickle, b16encode and base64

    :param data: any Python data object
    :returns: pickled string of data
    """
    try:
        return base64.b16encode(pickle.dumps(data))
    except:
        return ''

def decode_sdata(encoded_string):
    """
    Decode data pickled and encoded using encode_sdata

    :param encoded_string: pickled string of data
    :returns: unpickled data
    """
    try:
        return pickle.loads(base64.b16decode(encoded_string))
    except:
        return None


class DataObject(dict):
    """
    Data object class
    * based on webpy dict-like Storage object

    # DataObject class for storing generic dict key/value pairs
    # * object is a dict that behaves like a class object (key/value
    #   can be accessed like object attributes).
    #
    # borrowed from web.py
    #
    # class Storage(dict):
    #   '''
    #   A Storage object is like a dictionary except `obj.foo` can be used
    #   in addition to `obj['foo']`.
    #
    #       >>> o = storage(a=1)
    #       >>> o.a
    #       1
    #       >>> o['a']
    #       1
    #       >>> o.a = 2
    #       >>> o['a']
    #       2
    #       >>> del o.a
    #       >>> o.a
    #       Traceback (most recent call last):
    #           ...
    #       AttributeError: 'a'
    #
    #   '''
    #   def __getattr__(self, key):
    #       try:
    #           return self[key]
    #       except KeyError, k:
    #           raise AttributeError, k
    #
    #   def __setattr__(self, key, value):
    #       self[key] = value
    #
    #   def __delattr__(self, key):
    #       try:
    #           del self[key]
    #       except KeyError, k:
    #           raise AttributeError, k
    #
    #   def __repr__(self):
    #       return '<Storage ' + dict.__repr__(self) + '>'
    #
    """

    def __init__(self, *args, **kwargs):
        self.add(*args, **kwargs)

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError, k:
            raise AttributeError, k

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        try:
            del self[key]
        except KeyError, k:
            raise AttributeError, k

    def __repr__(self):
        return '<DataObject ' + dict.__repr__(self) + '>'

    def add(self, *args, **kwargs):
        """
        add({
            'a': 1,
            'b': 3.14
            'c': 'foo'
        })
        """
        for d in args:
            if isinstance(d, six.string_types):
                self[d] = True
            elif isinstance(d, dict):
                for name, value in d.items():
                    self[name] = value
            else:
                try:
                    for name in d:
                        self[name] = True
                except TypeError:
                    pass
        for name, value in kwargs.items():
            try:
                self[name] = value
            except TypeError:
                pass


def enum(*sequential, **named):
    """
    Definition for an `enum` type.

    # enum type
    #
    # from:
    # http://stackoverflow.com/questions/36932/whats-the-best-way-to-implement-an-enum-in-python
    #
    # def enum(**enums):
    #     return type('Enum', (), enums)
    # Used like so:
    #
    # >>> Numbers = enum(ONE=1, TWO=2, THREE='three')
    # >>> Numbers.ONE
    # 1
    # >>> Numbers.TWO
    # 2
    # >>> Numbers.THREE
    # 'three'
    # You can also easily support automatic enumeration with something like this:
    #
    # def enum(*sequential, **named):
    #     enums = dict(zip(sequential, range(len(sequential))), **named)
    #     return type('Enum', (), enums)
    # Used like so:
    #
    # >>> Numbers = enum('ZERO', 'ONE', 'TWO')
    # >>> Numbers.ZERO
    # 0
    # >>> Numbers.ONE
    # 1

    """
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type(str('Enum'), (), enums)


# get file extension

def get_file_ext(filename):
    """
    Extract extension for file.
    * This simple function is here for legacy reasons. Some projects
      import this function from this module so it's here to keep those
      from breaking.

    :param filename: filename or path
    :returns: tuple of file base name and extension (extension is '' if none)
    """
    return os.path.splitext(filename)


# for compatibiity with old versions

def cvt2list(s):
    """Convert object to list"""
    if not hasattr(s, '__iter__'):
        s = [s]
    return s


from garage.text_utils import (
    trim,
    check_eos,
    has_digits,
    has_alpha,
    has_alphanum,
    to_camel_case,
    substitute,
    subs,
    safe_unicode,
    safe_str
)



# -*- coding: utf-8 -*-
"""
garage.image_utils

Image-processing utility functions.
* Note: These functions read and write image files on a local file
  system.

* created: 2012-08-10 Kevin Chan <kefin@makedostudio.com>
* updated: 2014-11-21 kchan
"""

from __future__ import (absolute_import, unicode_literals)

import os.path
import re
import imghdr
try:
    from PIL import Image
except ImportError:
    import Image



# image utilities

DEFAULT_IMG_QUALITY = 75
RE_PATS = [
	r'^(.*)([_\-][0-9]+x[0-9]+)(\.[^\.]+)$',
	r'^(.*)(\.[^\.]+)$'
]
Regexp = None
DEFAULT_FNAME = 'image'


def resize_image(img, box, fit):
    """
    Downsample image and resize to 'box' dimensions.
    @param img: Image - an Image-object
    @param box: tuple(x, y) - the bounding box of the result image
    @param fit: boolean - crop the image to fill the box
    """
    w, h = box

    def f2i(n):
        return int(round(n))

    if fit:
        x, y = img.size
        src_ratio = 1.0 * x/y
        dst_ratio = 1.0 * w/h
        if src_ratio > dst_ratio:
            dy = 0.0
            dx = dst_ratio * y - x
        else:
            dx = 0.0
            dy = x / dst_ratio - y
        x += dx
        y += dy
        a0 = 1.0 * abs(dx/2)
        a1 = a0 + x
        b0 = 1.0 * abs(dy/2)
        b1 = b0 + y
        a0 = f2i(a0)
        a1 = f2i(a1)
        b0 = f2i(b0)
        b1 = f2i(b1)
        crop_coordinates = (a0, b0, a1, b1)
        img = img.crop(crop_coordinates)

    img = img.resize(box, Image.ANTIALIAS)
    return img


def get_image_size(image_file):
    img = Image.open(image_file)
    imgw, imgh = img.size
    return imgw, imgh


def create_thumb(image_file, w, h, quality, dst, fbase, fext):
    """
    Uses resize function above.
    """
    quality = int(quality)

    img = Image.open(image_file)

    imgw, imgh = img.size
    r = (1.0 * imgw) / imgh
    if not w:
        w = round(r * float(h))
    if not h:
        h = round(float(w) / r)

    w = int(w)
    h = int(h)

    img = resize_image(img, (w, h,), True)
    if img.mode != 'RGB':
        img = img.convert('RGB')

    filename = '%s-%dx%d.%s' % (fbase, w, h, fext)
    output = os.path.join(dst, filename)
    img.save(output, quality=quality)
    return output


def get_file_basename(f, default=DEFAULT_FNAME):
    global Regexp
    if Regexp is None:
        Regexp = [re.compile(r, re.I) for r in RE_PATS]
    fname = os.path.basename(f)
    for r in Regexp:
        m = r.match(fname)
        if m:
            return m.group(1)
    return default


def get_img_ext(path, default_ext='unknown'):
    """
    Detect image type from file path and return file extension.
    """
    imgtype = imghdr.what(path)
    suffix = {
        'rgb': 'rgb',
        'gif': 'gif',
        'pbm': 'pbm',
        'pgm': 'pgm',
        'ppm': 'ppm',
        'tiff': 'tif',
        'rast': 'rast',
        'xbm': 'xbm',
        'jpeg': 'jpg',
        'bmp': 'bmp',
        'png': 'png'
        }
    return suffix.get(imgtype, default_ext)


def generate_thumb(img_file, width, height, quality, dest_dir):
    """
    Process image file and create thumbnail according to parameters.
    """
    fbase = get_file_basename(img_file)
    fext = get_img_ext(img_file)
    return create_thumb(img_file, width, height, quality, dest_dir, fbase, fext)
